==> hotplug-diskmount-0.2.tar.gz: hotplug-diskmount-0.2/hotplug-diskmount.c <==
/*
 * Written by Alexey Vatchenko <av@bsdua.org>.
 * Public domain.
 */
#include <sys/types.h>
#include <sys/param.h>

#include <sys/disklabel.h>
#include <sys/dkio.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define VOL_ROOT		"/vol"
#define VOL_DB			VOL_ROOT "/.db"
#define DEV_ROOT		"/dev"
#define DISKNAME_SIZE		16
#define CHECK_MEDIA_TIMEOUT	2

static uid_t mp_uid = 0;	/* default owner of mountpoint */
static gid_t mp_gid = 0;	/* default group of mountpoint */
/* default access permissions of mountpoint */
static mode_t mp_perm = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
static int ignore_umask = 0;

void	usage(const char *prog);
int	get_disk_label(const char *devname, char **dev,
	    struct disklabel *label);
void	prepare_disk_name(char *dskname, struct disklabel *label);
int	compose_mp(const char *dskname, char **mp, int multi, int cnt);
int	create_mp(const char *dskname, char **mp, int multi, int cnt);
int	save_mp_info(const char *devname, int pi, const char *mp);
int	do_mount(const char *dev, const char *mp);
int	mount_disk(const char *devname, char *dev, const char *dskname,
	    struct disklabel *label);
int	perm_str2val(const char *str, mode_t *mode);
int	remove_mount_points(const char *devname);

void
usage(const char *prog)
{
	printf("Usage:\n"
	    "\t%s attach [-u user] [-g group] [-m mode] device\n"
	    "\t%s cleanup device\n", prog, prog);
}

int
get_disk_label(const char *devname, char **dev, struct disklabel *label)
{
	int fd;
	int error, rv;
	size_t sz;

	sz = sizeof(DEV_ROOT) + strlen(devname) + 1 + 1;
	*dev = (char *)malloc(sz);
	if (*dev == NULL)
		return (ENOMEM);
	strlcpy(*dev, DEV_ROOT, sz);
	strlcat(*dev, "/", sz);
	strlcat(*dev, devname, sz);
	strlcat(*dev, "c", sz);

	fd = open(*dev, O_RDONLY);
	rv = errno;
	if (fd == -1) {
		free(*dev);
		return (rv);
	}

	error = ioctl(fd, DIOCGDINFO, label);
	rv = errno;
	close(fd);
	if (error == -1) {
		free(*dev);
		return (rv);
	}

	return (0);
}

void
prepare_disk_name(char *dskname, struct disklabel *label)
{
	int i, j, st, sz;

	/* Skip leading spaces */
	for (i = 0; i < DISKNAME_SIZE - 1; i++) {
		if (!isspace(label->d_un.un_d_packname[i]))
			break;
	}

	/* Skip trailing spaces */
	st = i;
	for (i = DISKNAME_SIZE - 1; i > st; i--) {
		if (!isspace(label->d_un.un_d_packname[i - 1]))
			break;
	}
	sz = i;

	for (i = st, j = 0; i < sz; i++, j++) {
		if (label->d_un.un_d_packname[i] == '/')
			dskname[j] = '_';
		else
			dskname[j] = label->d_un.un_d_packname[i];
	}
	dskname[j] = '\0';

	if (*dskname == '\0')
		strlcpy(dskname, "NONAME", DISKNAME_SIZE);
	if (*dskname == '.')
		*dskname = '_';	/* unhide mount points */
}

int
compose_mp(const char *dskname, char **mp, int multi, int cnt)
{
	size_t sz, dsklen, cntsz;
	int i, div;

	dsklen = strlen(dskname);
	sz = sizeof(VOL_ROOT) + dsklen + 1;
	if (multi) {
		/* /vol/NameOfDisk-p1 */
		cntsz = 1;
		div = cnt;
		while ((div /= 10) > 0)
			cntsz++;

		sz += cntsz + 2;
	}

	*mp = (char *)malloc(sz);
	if (*mp == NULL)
		return (ENOMEM);

	strlcpy(*mp, VOL_ROOT, sz);
	strlcat(*mp, "/", sz);
	strlcat(*mp, dskname, sz);

	if (multi) {
		strlcat(*mp, "-p", sz);
		div = cnt;
		for (i = sz - 2; i > sz - cntsz - 2; i--) {
			(*mp)[i] = '0' + div % 10;
			div /= 10;
		}
	}

	return (0);
}

int
create_mp(const char *dskname, char **mp, int multi, int cnt)
{
	int error;
	int i;
	char *p;
	size_t mpsz;
	mode_t old_umask;
	int old_errno;

	error = compose_mp(dskname, mp, multi, cnt);
	if (error != 0)
		return (error);

	mpsz = strlen(*mp);
	p = NULL;

	for (i = 1; i < 1000; i++) {
		if (ignore_umask)
			old_umask = umask(0);
		error = mkdir(*mp, mp_perm);
		old_errno = errno;
		if (error != -1)
			chown(*mp, mp_uid, mp_gid);
		if (ignore_umask)
			umask(old_umask);
		errno = old_errno;

		if (error != -1 || errno != EEXIST) {
			error = (error == -1) ? errno : 0;
			break;
		}

		if (p == NULL) {
			/* ' #NNN' */
			p = (char *)realloc(*mp, mpsz + 5 + 1);
			if (p == NULL) {
				error = ENOMEM;
				break;
			}
		}

		snprintf(p + mpsz, 5 + 1, " #%u", i);
		*mp = p;

		error = EEXIST;
	}

	if (error != 0)
		free(*mp);
	return (error);
}

int
save_mp_info(const char *devname, int pi, const char *mp)
{
	char *lnk;
	size_t sz;
	int error, rv;

	sz = sizeof(VOL_DB) + strlen(devname) + 1 + 1;
	lnk = (char *)malloc(sz);
	if (lnk == NULL)
		return (ENOMEM);
	snprintf(lnk, sz, "%s/%s%c", VOL_DB, devname, 'a' + pi);
	unlink(lnk);
	error = symlink(mp, lnk);
	error = (error == -1) ? errno : 0;
	free(lnk);
	return (rv);
}

int
do_mount(const char *dev, const char *mp)
{
	int pid;
	const char *argv[] = {
	    "mount", "-onodev,nosuid", NULL, NULL, NULL};

	argv[2] = dev;
	argv[3] = mp;
	pid = fork();
	if (pid == 0) {
		setsid();
		close(0);
		close(1);
		close(2);
		execv("/sbin/mount", argv);
		exit(errno);
	}
	return ((pid > 0) ? 0 : errno);
}

int
mount_disk(const char *devname, char *dev, const char *dskname,
    struct disklabel *label)
{
	int i, np, pi;
	char *mp, *dev_part;
	int ignore_c;
	int error;

	np = 0;
	for (i = 0; i < label->d_npartitions; i++) {
		if (label->d_partitions[i].p_fstype != FS_UNUSED)
			np++;
	}

	if (np == 0)
		return (0);

	/*
	 * Sometimes, "c" partition points to the same data as another
	 * partition. So we should exclude "c" partition.
	 */
	ignore_c = 0;
	if (np > 1 && label->d_npartitions > 2 &&
	    label->d_partitions[2].p_fstype != FS_UNUSED) {
		ignore_c = 1;
		np--;
	}

	dev_part = dev + strlen(dev) - 1;
	pi = 1;
	for (i = 0; i < label->d_npartitions; i++) {
		if (label->d_partitions[i].p_fstype != FS_UNUSED &&
		    !(i == 2 && ignore_c)) {
			error = create_mp(dskname, &mp, (np > 1), pi);
			if (error == 0) {
				save_mp_info(devname, i, mp);
				*dev_part = 'a' + i;
				do_mount(dev, mp);
				free(mp);
			}
			pi++;
		}
	}

	return (0);
}

int
perm_str2val(const char *str, mode_t *mode)
{
	size_t i, len;
	mode_t md;

	len = strlen(str);
	if (len != 3 && len != 4)
		return (-1);

	md = 0;
	for (i = len; i > 0; i--) {
		if (str[i - 1] < '0' || str[i - 1] > '7')
			return (-1);
		md |= (str[i - 1] - '0') << (3 * (len - i));
	}

	*mode = md;
	return (0);
}

int
remove_mount_points(const char *devname)
{
	char mp[MAXPATHLEN + 1];
	char *lnk;
	size_t sz;
	int i;
	int error;

	sz = sizeof(VOL_DB) + strlen(devname) + 1 + 1;
	lnk = (char *)malloc(sz);
	if (lnk == NULL)
		return (ENOMEM);
	snprintf(lnk, sz, "%s/%s%c", VOL_DB, devname, 'a');

	for (i = 0; i < MAXPARTITIONS; i++) {
		lnk[sz - 2] = 'a' + i;
		memset(mp, 0, sizeof(mp));
		error = readlink(lnk, mp, sizeof(mp) - 1);
		if (error != -1)
			rmdir(mp);
		unlink(lnk);
	}

	return (0);
}

int
main(int argc, const char **argv)
{
	struct disklabel label;
	const char *arg_cmd, *arg_devname;
	const char *progname;
	char diskname[16], *dev;
	struct passwd *pwent;
	struct group *grent;
	int succeed, error;
	int ch;
	mode_t mode;
	pid_t pid;

	arg_devname = NULL;
	progname = argv[0];

	if (argc < 2) {
		usage(progname);
		return (1);
	}

	arg_cmd = argv[1];

	if (strcmp(arg_cmd, "attach") == 0) {
		while ((ch = getopt(argc - 1, argv + 1, "g:m:u:")) != -1) {
			switch (ch) {
			case 'g':
				grent = getgrnam(optarg);
				if (grent == NULL) {
					fprintf(stderr,
					    "specified group is not found\n");
					exit(1);
				}
				mp_gid = grent->gr_gid;
				break;
			case 'm':
				error = perm_str2val(optarg, &mode);
				if (error != 0) {
					fprintf(stderr,
					    "invalid mode is specified\n");
					exit(1);
				}
				mp_perm = mode;
				ignore_umask = 1;
				break;
			case 'u':
				pwent = getpwnam(optarg);
				if (pwent == NULL) {
					fprintf(stderr,
					    "specified user is not found\n");
					exit(1);
				}
				mp_uid = pwent->pw_uid;
				break;
			default:
				usage(progname);
				exit(1);
			}
		}

		argc -= optind + 1;
		argv += optind + 1;

		if (argc < 1) {
			usage(progname);
			exit(1);
		}

		arg_devname = argv[0];
	} else if (strcmp(arg_cmd, "cleanup") == 0) {
		if (argc < 3) {
			usage(progname);
			exit(1);
		}

		arg_devname = argv[2];
	} else {
		usage(progname);
		exit(1);
	}

	if (strcmp(arg_cmd, "attach") == 0) {
		signal(SIGCHLD, SIG_IGN);

		pid = fork();
		if (pid != 0)
			exit(0);
		setsid();
		close(0);
		close(1);
		close(2);

		succeed = 0;
		for (;;) {
			error = get_disk_label(arg_devname, &dev, &label);
			if (error == ENOMEDIUM) {
				/*
				 * If the media were removed from the
				 * device, cleanup mount points.
				 */
				if (succeed) {
					remove_mount_points(arg_devname);
					succeed = 0;
				}
			} else if (error == EBUSY) {
				/*
				 * The device is mounted.
				 */
				/* NOTHING */
			} else if (error == 0 && succeed) {
				/*
				 * Someone umounted CD/DVD...
				 * wait for eject/insert.
				 */
				/* NOTHING */
			} else if (error == 0) {
				prepare_disk_name(diskname, &label);
				error = mount_disk(arg_devname, dev, diskname,
				    &label);
				free(dev);
				succeed = 1;
			} else {
				if (succeed)
					remove_mount_points(arg_devname);
				exit(0);
			}

			sleep(CHECK_MEDIA_TIMEOUT);
		}
		/* NOTREACHED */
	} else if (strcmp(arg_cmd, "cleanup") == 0) {
		error = remove_mount_points(arg_devname);
		exit(error != 0);
	}

	usage(progname);
	exit(1);
}

==> hotplug-diskmount-0.4.tar.gz: hotplug-diskmount-0.4/hotplug-diskmount.c <==
/*
 * Written by Alexey Vatchenko <av@bsdua.org>.
 * Public domain.
 */
#include <sys/types.h>
#include <sys/param.h>

#include <sys/disklabel.h>
#include <sys/dkio.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <sys/stat.h>

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define DEFAULT_VOL_ROOT	"/vol"
#define DEFAULT_VOL_DB		"/.db"
#define DEV_ROOT		"/dev"
#define DISKNAME_SIZE		16
#define CHECK_MEDIA_TIMEOUT	2

static uid_t mp_uid = 0;	/* default owner of mountpoint */
static gid_t mp_gid = 0;	/* default group of mountpoint */
/* default access permissions of mountpoint */
#define DEFAULT_PERM	(S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)
static mode_t mp_perm = DEFAULT_PERM;
static int ignore_umask = 0;
static volatile sig_atomic_t quit_flag = 0;
static char vol_root[MAXPATHLEN] = DEFAULT_VOL_ROOT;
static char vol_db[MAXPATHLEN] = DEFAULT_VOL_ROOT DEFAULT_VOL_DB;

void	usage(const char *prog);
void	sig_term(int sig);
int	get_disk_label(const char *devname, char **dev,
	    struct disklabel *label);
void	prepare_disk_name(char *dskname, struct disklabel *label);
int	compose_mp(const char *dskname, char **mp, int multi, int cnt);
int	create_mp(const char *dskname, char **mp, int multi, int cnt);
int	save_mp_info(const char *devname, int pi, const char *mp);
int	do_mount(const char *dev, const char *mp);
void	mount_disk(const char *devname, char *dev, const char *dskname,
	    struct disklabel *label);
int	perm_str2val(const char *str, mode_t *mode);
int	remove_mount_points(const char *devname, int umnt);
int	create_root(void);

void
usage(const char *prog)
{
	printf("Usage:\n"
	    "\t%s [-d dir] attach [-u user] [-g group] [-m mode] device\n"
	    "\t%s [-d dir] init\n"
	    "\t%s [-d dir] cleanup device\n", prog, prog, prog);
}

void
sig_term(int sig)
{
	quit_flag = 1;
}

int
get_disk_label(const char *devname, char **dev, struct disklabel *label)
{
	int fd;
	int error, rv;
	size_t sz;

	sz = sizeof(DEV_ROOT) + strlen(devname) + 1 + 1;
	*dev = (char *)malloc(sz);
	if (*dev == NULL)
		return (ENOMEM);
	strlcpy(*dev, DEV_ROOT, sz);
	strlcat(*dev, "/", sz);
	strlcat(*dev, devname, sz);
	strlcat(*dev, "c", sz);

	fd = open(*dev, O_RDONLY);
	rv = errno;
	if (fd == -1) {
		free(*dev);
		return (rv);
	}

	error = ioctl(fd, DIOCGDINFO, label);
	rv = errno;
	close(fd);
	if (error == -1) {
		free(*dev);
		return (rv);
	}

	return (0);
}

void
prepare_disk_name(char *dskname, struct disklabel *label)
{
	int i, j, st, sz;

	/* Skip leading spaces */
	for (i = 0; i < DISKNAME_SIZE - 1; i++) {
		if (!isspace(label->d_un.un_d_packname[i]))
			break;
	}

	/* Skip trailing spaces */
	st = i;
	for (i = DISKNAME_SIZE - 1; i > st; i--) {
		if (!isspace(label->d_un.un_d_packname[i - 1]) &&
		    label->d_un.un_d_packname[i - 1] != '\0')
			break;
	}
	sz = i;

	for (i = st, j = 0; i < sz; i++, j++) {
		if (label->d_un.un_d_packname[i] == '/')
			dskname[j] = '_';
		else
			dskname[j] = label->d_un.un_d_packname[i];
	}
	dskname[j] = '\0';

	if (*dskname == '\0')
		strlcpy(dskname, "NONAME", DISKNAME_SIZE);
	if (*dskname == '.')
		*dskname = '_';	/* unhide mount points */
}

int
compose_mp(const char *dskname, char **mp, int multi, int cnt)
{
	size_t sz, dsklen, cntsz;
	int i, div;

	dsklen = strlen(dskname);
	sz = strlen(vol_root) + 1 + dsklen + 1;
	if (multi) {
		/* /vol/NameOfDisk-p1 */
		cntsz = 1;
		div = cnt;
		while ((div /= 10) > 0)
			cntsz++;

		sz += cntsz + 2;
	}

	*mp = (char *)malloc(sz);
	if (*mp == NULL)
		return (ENOMEM);

	strlcpy(*mp, vol_root, sz);
	strlcat(*mp, "/", sz);
	strlcat(*mp, dskname, sz);

	if (multi) {
		strlcat(*mp, "-p", sz);
		div = cnt;
		for (i = sz - 2; i > sz - cntsz - 2; i--) {
			(*mp)[i] = '0' + div % 10;
			div /= 10;
		}
	}

	return (0);
}

int
create_mp(const char *dskname, char **mp, int multi, int cnt)
{
	int error;
	int i;
	char *p;
	size_t mpsz;
	mode_t old_umask;
	int old_errno;

	error = compose_mp(dskname, mp, multi, cnt);
	if (error != 0)
		return (error);

	mpsz = strlen(*mp);
	p = NULL;

	for (i = 1; i < 1000; i++) {
		if (ignore_umask)
			old_umask = umask(0);

		error = mkdir(*mp, mp_perm);
		old_errno = errno;

		if (error == -1 && errno == EEXIST) {
			error = rmdir(*mp);
			if (error == 0) {
				error = mkdir(*mp, mp_perm);
				old_errno = errno;
			}
		}
		if (error != -1)
			chown(*mp, mp_uid, mp_gid);

		if (ignore_umask)
			umask(old_umask);

		errno = old_errno;
		if (error != -1 || errno != EEXIST) {
			error = (error == -1) ? errno : 0;
			break;
		}

		if (p == NULL) {
			/* ' #NNN' */
			p = (char *)realloc(*mp, mpsz + 5 + 1);
			if (p == NULL) {
				error = ENOMEM;
				break;
			}
		}

		snprintf(p + mpsz, 5 + 1, " #%u", i);
		*mp = p;

		error = EEXIST;
	}

	if (error != 0)
		free(*mp);
	return (error);
}

int
save_mp_info(const char *devname, int pi, const char *mp)
{
	char *lnk;
	size_t sz;
	int error, rv;

	sz = strlen(vol_db) + 1 + strlen(devname) + 1 + 1;
	lnk = (char *)malloc(sz);
	if (lnk == NULL)
		return (ENOMEM);
	snprintf(lnk, sz, "%s/%s%c", vol_db, devname, 'a' + pi);
	unlink(lnk);
	error = symlink(mp, lnk);
	rv = (error == -1) ? errno : 0;
	free(lnk);
	return (rv);
}

int
do_mount(const char *dev, const char *mp)
{
	int pid;
	const char *argv[] = {
	    "mount", "-onodev,nosuid", NULL, NULL, NULL};

	argv[2] = dev;
	argv[3] = mp;
	pid = fork();
	if (pid == 0) {
		setsid();
		close(0);
		close(1);
		close(2);
		execv("/sbin/mount", argv);
		exit(errno);
	}
	return ((pid > 0) ? 0 : errno);
}

void
mount_disk(const char *devname, char *dev, const char *dskname,
    struct disklabel *label)
{
	int i, np, pi;
	char *mp, *dev_part;
	int ignore_c;
	int error;

	np = 0;
	for (i = 0; i < label->d_npartitions; i++) {
		if (label->d_partitions[i].p_fstype != FS_UNUSED)
			np++;
	}

	if (np == 0)
		return;

	/*
	 * Sometimes, "c" partition points to the same data as another
	 * partition. So we should exclude "c" partition.
	 */
	ignore_c = 0;
	if (np > 1 && label->d_npartitions > 2 &&
	    label->d_partitions[2].p_fstype != FS_UNUSED) {
		ignore_c = 1;
		np--;
	}

	dev_part = dev + strlen(dev) - 1;
	pi = 1;
	for (i = 0; i < label->d_npartitions; i++) {
		if (label->d_partitions[i].p_fstype != FS_UNUSED &&
		    !(i == 2 && ignore_c)) {
			error = create_mp(dskname, &mp, (np > 1), pi);
			if (error == 0) {
				save_mp_info(devname, i, mp);
				*dev_part = 'a' + i;
				do_mount(dev, mp);
				free(mp);
			}
			pi++;
		}
	}
}

int
perm_str2val(const char *str, mode_t *mode)
{
	size_t i, len;
	mode_t md;

	len = strlen(str);
	if (len != 3 && len != 4)
		return (-1);

	md = 0;
	for (i = len; i > 0; i--) {
		if (str[i - 1] < '0' || str[i - 1] > '7')
			return (-1);
		md |= (str[i - 1] - '0') << (3 * (len - i));
	}

	*mode = md;
	return (0);
}

int
remove_mount_points(const char *devname, int umnt)
{
	char mp[MAXPATHLEN + 1];
	char *lnk;
	size_t sz;
	int i;
	int error;

	sz = strlen(vol_db) + 1 + strlen(devname) + 1 + 1;
	lnk = (char *)malloc(sz);
	if (lnk == NULL)
		return (ENOMEM);
	snprintf(lnk, sz, "%s/%s%c", vol_db, devname, 'a');

	for (i = 0; i < MAXPARTITIONS; i++) {
		lnk[sz - 2] = 'a' + i;
		memset(mp, 0, sizeof(mp));
		error = readlink(lnk, mp, sizeof(mp) - 1);
		if (error != -1) {
			if (umnt) {
				error = unmount(mp, 0);
				if (error == -1 && errno == EBUSY)
					return (EBUSY);
			}
			rmdir(mp);
		}
		unlink(lnk);
	}

	return (0);
}

int
create_root(void)
{
	int error, saved_errno;
	char *errstr;

	error = mkdir(vol_root, DEFAULT_PERM);
	if (error != 0 && errno != EEXIST) {
		saved_errno = errno;
		errstr = strerror(saved_errno);
		fprintf(stderr, "Cannot create directory: %s\n", errstr);
		return (saved_errno);
	}

	error = mkdir(vol_db, S_IRWXU);
	if (error != 0 && errno != EEXIST) {
		saved_errno = errno;
		errstr = strerror(saved_errno);
		fprintf(stderr, "Cannot create directory: %s\n", errstr);
		return (saved_errno);
	}

	return (0);
}

int
main(int argc, char * const *argv)
{
	struct disklabel label;
	const char *arg_cmd, *arg_devname;
	const char *progname;
	char diskname[16], *dev;
	struct passwd *pwent;
	struct group *grent;
	int succeed, error;
	int i, ch;
	mode_t mode;
	pid_t pid;
	size_t len;

	arg_devname = NULL;
	progname = argv[0];

	while ((ch = getopt(argc, argv, "d:")) != -1) {
		switch (ch) {
		case 'd':
			len = strlcpy(vol_root, optarg, sizeof(vol_root));
			if (len >= sizeof(vol_root)) {
				fprintf(stderr, "%s: path is too long\n",
				    progname);
				exit(1);
			}

			/*
			 * Remove trailing slash but allow user to specify root
			 * direcotry "/".
			 */
			for (i = len; i > 1; i--) {
				if (vol_root[i - 1] == '/')
					vol_root[i - 1] = '\0';
				else
					break;
			}

			/*
			 * Empty string is not allowed. Otherwise, mount points
			 * will be created in root diriectory.
			 */
			if (*vol_root == '\0') {
				fprintf(stderr, "%s: path is invalid\n",
				    progname);
				exit(1);
			}

			/* change vol_db path */
			strlcpy(vol_db, vol_root, sizeof(vol_db));
			len = strlcat(vol_db, DEFAULT_VOL_DB, sizeof(vol_db));
			if (len >= sizeof(vol_db)) {
				fprintf(stderr, "%s: path is too long\n",
				    progname);
				exit(1);
			}

			break;
		default:
			usage(progname);
			exit(1);
		}
	}

	argc -= optind;
	argv += optind;
	optind = 0;
	optreset = 1;

	if (argc < 1) {
		usage(progname);
		return (1);
	}

	arg_cmd = argv[0];

	if (strcmp(arg_cmd, "attach") == 0) {
		while ((ch = getopt(argc, argv, "g:m:u:")) != -1) {
			switch (ch) {
			case 'g':
				grent = getgrnam(optarg);
				if (grent == NULL) {
					fprintf(stderr, "%s: "
					    "specified group is not found\n",
					    progname);
					exit(1);
				}
				mp_gid = grent->gr_gid;
				break;
			case 'm':
				error = perm_str2val(optarg, &mode);
				if (error != 0) {
					fprintf(stderr,
					    "%s: invalid mode is specified\n",
					    progname);
					exit(1);
				}
				mp_perm = mode;
				ignore_umask = 1;
				break;
			case 'u':
				pwent = getpwnam(optarg);
				if (pwent == NULL) {
					fprintf(stderr,
					    "%s: specified user is not found\n",
					    progname);
					exit(1);
				}
				mp_uid = pwent->pw_uid;
				break;
			default:
				usage(progname);
				exit(1);
			}
		}

		argc -= optind;
		argv += optind;

		if (argc < 1) {
			usage(progname);
			exit(1);
		}

		arg_devname = argv[0];
	} else if (strcmp(arg_cmd, "init") == 0) {
		/* nothing */
	} else if (strcmp(arg_cmd, "cleanup") == 0) {
		if (argc < 2) {
			usage(progname);
			exit(1);
		}

		arg_devname = argv[1];
	} else {
		usage(progname);
		exit(1);
	}

	if (strcmp(arg_cmd, "attach") == 0) {
		signal(SIGCHLD, SIG_IGN);

		pid = fork();
		if (pid != 0)
			exit(0);
		setsid();
		close(0);
		close(1);
		close(2);

		signal(SIGTERM, sig_term);
		succeed = 0;
		for (;;) {
			if (quit_flag) {
				if (succeed) {
					for (i = 0; i < 5; i++) {
						error = remove_mount_points(
						    arg_devname, 1);
						if (error != EBUSY)
							break;
						sleep(1);
					}
				}
				exit(0);
			}

			error = get_disk_label(arg_devname, &dev, &label);
			if (error == ENOMEDIUM) {
				/*
				 * If the media were removed from the
				 * device, cleanup mount points.
				 */
				if (succeed) {
					remove_mount_points(arg_devname, 0);
					succeed = 0;
				}
			} else if (error == EBUSY) {
				/*
				 * The device is mounted.
				 */
				/* NOTHING */
			} else if (error == 0 && succeed) {
				/*
				 * Someone umounted CD/DVD...
				 * wait for eject/insert.
				 */
				/* NOTHING */
			} else if (error == 0) {
				prepare_disk_name(diskname, &label);
				mount_disk(arg_devname, dev, diskname, &label);
				free(dev);
				succeed = 1;
			} else {
				if (succeed)
					remove_mount_points(arg_devname, 0);
				exit(0);
			}

			sleep(CHECK_MEDIA_TIMEOUT);
		}
		/* NOTREACHED */
	} else if (strcmp(arg_cmd, "init") == 0) {
		error = create_root();
		exit(error != 0);
	} else if (strcmp(arg_cmd, "cleanup") == 0) {
		error = remove_mount_points(arg_devname, 0);
		exit(error != 0);
	}

	usage(progname);
	exit(1);
}
